%% INPAINT_MMSE - Image inpainting by approximating the MMSE estimate with samples
% |IMG_INPAINTED = INPAINT_MMSE(MRF, IMG, MASK[, RB, DOPLOT])| will use MRF to inpaint
% the pixels of IMG that are specified by MASK (boolean matrix of same size as IMG).
% IMG can either be a gray-scale or RGB image.
% Set RB to true to use a Rao-Blackwellized MMSE estimator (not used in the paper, suggested by George Papandreou).
% Set DOPLOT to true if you want to see the inpainting progress.
% Optional arguments default to false.
% 
% This file is part of the implementation as described in the paper:
% 
%  Uwe Schmidt, Qi Gao, Stefan Roth.
%  A Generative Perspective on MRFs in Low-Level Vision.
%  IEEE Conference on Computer Vision and Pattern Recognition (CVPR'10), San Francisco, USA, June 2010.
% 
% Please cite the paper if you are using this code in your work.
% 
% The code may be used free of charge for non-commercial and
% educational purposes, the only requirement is that this text is
% preserved within the derivative work. For any other purpose you
% must contact the authors for permission. This code may not be
% redistributed without permission from the authors.
%
%  Author:  Uwe Schmidt, Department of Computer Science, TU Darmstadt
%  Contact: mail@uweschmidt.org
% 
% Project page:  http://www.gris.tu-darmstadt.de/research/visinf/software/index.en.htm

% Copyright 2009-2011 TU Darmstadt, Darmstadt, Germany.
% $Id: inpaint_mmse.m 238 2011-05-23 14:11:56Z uschmidt $

function img_inpainted = inpaint_mmse(mrf, img, mask, doplot)
  
  if ~isa(mrf, 'pml.distributions.gsm_foe'), error('MRF unsuitable'), end
  
  mrf.update_filter_matrices = true;
  mrf.conv_method = 'valid';
  
  nsamplers = 3;
  max_iters = ceil(1800 / nsamplers); % use at most this many samples to obtain the inpainting image
  max_burnin_iters = 100; % only the second half of these will be considered (i.e. throw away 50 initial samples at max)
  mean_mapd_threshold = 1; % convergence threshold
  
  max_mapd_diff_threshold = 0.1; % convergence threshold
  min_nsamples = 40;
  
  discard_percent = 2/3; % percent of the samples to discard
  
  % always use Rao-Blackwellization!
  
  % plot while inpainting?
  if ~exist('doplot', 'var'), doplot = false; end
  if doplot, figure(1), clf, end
  
  nlayers = size(img, 3);
  img_corrupted = img;
    
  % assuming RGB image, convert to YCbCr
  if nlayers == 3, img = 255 * rgb2ycbcr(double(img) / 255); end
  
  I = img;
  img_inpainted = img;
  
  mrf.imdims = size(I(:,:,1));
  npixels = numel(I(:,:,1));
  ind_int = setdiff((1:npixels)' .* mask(:), 0); % pixels to be inpainted (ones in mask)
  ind_ext = setdiff(1:npixels, ind_int);         % "support" pixels (zeros in mask)
  
  img_vector = reshape(I, npixels, nlayers);
  
  % init all samplers with the image
  x = repmat(img_vector, [1, 1, nsamplers]);
  
  % overwrite first 2 samplers with simple "inpainted" images (smooth & noisy)
  for j = 1:nlayers
    x(:,j,1) = reshape(255 * imfilter(I(:,:,j)/255, fspecial('gaussian', 10, 5)), npixels, 1);
    x(ind_ext,j,1) = img_vector(ind_ext,j);
    x(ind_int,j,2) = img_vector(ind_int,j) + 10 * randn(numel(ind_int), 1);
  end
    
  % to store inpainted image under each sampler
  x_inpainted = zeros(npixels, nlayers, nsamplers);
  % to store the means for the Gaussian distributions (for Rao-Blackwellization)
  x_mu = zeros(npixels, nlayers, nsamplers);
  
  mapd = zeros(nsamplers, max_iters);
  
  % save all samples of the burn-in
  x_burnin = zeros(npixels, nlayers, nsamplers, max_burnin_iters);
  
  % only one scalar estimand for the energy
  estimands = zeros(max_burnin_iters, nsamplers, nlayers);
  R_hat = zeros(nlayers, max_burnin_iters);
  
  iter = 0; c = 0;
  burnin = true;
    
  % loop until convergence or max_iters depleted
  while true
    iter = iter + 1;
    
    % advance all samplers
    for i = 1:nsamplers
      for j = 1:nlayers
        z = mrf.sample_z(x(:,j,i));
        [x(:,j,i), x_mu(:,j,i)] = sample_x_inpainting(mrf, z, mask(:), img_vector(:,j));
        
        % save all samples in the burn-in phase
        if burnin
          x_burnin(:,j,i,iter) = x_mu(:,j,i);
          % average estimands in case of color image
          estimands(iter,i,j) = mrf.energy(x_mu(:,j,i));
        end
      end
    end
        
    % Rao-Blackwellized estimator, which can lead to faster convergence.
    sample = x_mu;
    
    if doplot && false
        figure(2);
        clf;
        subplot(131), imshow(uint8(reshape(sample(:,:,1), [mrf.imdims, nlayers])), 'InitialMagnification', 'fit'), title 'blur'
        subplot(132), imshow(uint8(reshape(sample(:,:,2), [mrf.imdims, nlayers])), 'InitialMagnification', 'fit'), title 'noise'
        subplot(133), imshow(uint8(reshape(sample(:,:,3), [mrf.imdims, nlayers])), 'InitialMagnification', 'fit'), title 'orig'
        drawnow;
        figure(1);
    end
    
    if doplot
        %subplot: y, x, n
        subplot(231), imshow(uint8(img_corrupted), 'InitialMagnification', 'fit'), title 'Initial'
        subplot(232), imshow(uint8(img_inpainted), 'InitialMagnification', 'fit'), title 'Inpainted'
        subplot(233), imshow(uint8(reshape(mean(sample, 3), [mrf.imdims, nlayers])), 'InitialMagnification', 'fit'), title 'Sample'
        
        % does not work this way. see sample_z documentation. is there a
        % plot_z function??
        if false
            figure(2)
            z_img = 0;
            if length(z) == 2
                z1 = [reshape(z{1}, [mrf.imdims(1), mrf.imdims(2)-1]) zeros(171,1)];
                z2 = [reshape(z{2}, [mrf.imdims(1)-1, mrf.imdims(2)]); zeros(1, 171)];
                z_img = log(z1) + log(z2);
            else
                z_img = zeros([mrf.imdims(1)-2, mrf.imdims(2)-2]);
                for i = 1:length(z)
                    z_img = z_img + reshape(log(z{i}), [mrf.imdims(1)-2, mrf.imdims(2)-2]);
                end
            end
            subplot(111), imshow(-z_img, [],  'InitialMagnification', 'fit'), title 'Z'
            %subplot(212), imshow(reshape(z{2}, [mrf.imdims(1)-1, mrf.imdims(2)]), [],  'InitialMagnification', 'fit'), title 'Z'
        end
        drawnow;
        figure(1)
    end
    
    % check for convergence in burn-in phase
    if burnin
      % compute epsr, ignoring first half of samples
      R_hat(:, iter) = pml.support.epsr(estimands(ceil(iter*discard_percent):iter,:,:));
      fprintf('\rBurn-in %2d / %2d, R_hat = %.3f', iter, max_burnin_iters, mean(R_hat(:,iter)))
      
      if doplot
        img_inpainted = reshape(mean(mean(x_burnin(:, :, :, ceil(iter*discard_percent):iter), 4),3), [mrf.imdims, nlayers]);
        
        %subplot(211), plot(1:iter, R_hat(:,1:iter)), title 'R\_hat (estimated potential scale reduction)'
        subplot(212), title 'Energies of burn-in samples' % blue, green, red -> blur, noise, orig
        for j = 1:nlayers, plot(1:iter, estimands(1:iter,:,j)), hold on, end
        drawnow
      end
      
      % discard at least 5 samples
      if (iter >= 10) && (all(R_hat(:,iter) < 1.1) || (iter > max_burnin_iters))
        if doplot
            subplot(212), clf;
            img_corrupted = img_inpainted;
        end
          
        burnin = false;
        % use second half of samples to compute inpainted images under each sampler
        x_inpainted = mean(x_burnin(:, :, :, ceil(iter*discard_percent):iter), 4);
        
        % set counter c to enable running average
        c = length(ceil(iter*discard_percent):iter);
        
        mapd(:,1:c) = nan;
        
        if iter > max_burnin_iters
          warning('Didn''t reach convergence in burn-in phase.')
        end
      end
    % after burn-in phase
    else
      % update inpainted images under each sampler (running average)
      x_inpainted = (c * x_inpainted + sample) / (c + 1);
      c = c + 1;
      
      % overall inpainted image as average of all samplers
      x_final = mean(x_inpainted, 3);
      
      img_inpainted = reshape(x_final, [mrf.imdims, nlayers]);
      if nlayers == 3, img_inpainted = 255 * ycbcr2rgb(img_inpainted / 255); end
      
      % (averaged over layers) mean absolute pixel distance (considering only inpainted pixels)
      mapd(:,c) = mean(mean(abs(bsxfun(@minus, x_inpainted(ind_int,:,:), x_final(ind_int,:)))));
      
      % display current status
      fprintf('\rInpainting :: sample %3d / %3d, avg. meanAPD = %7.4f', ...
              c, max_iters, mean(mapd(:,c)))
  
      if doplot
        subplot(212), plot(1:c, mapd(:,1:c)', 1:c, mean(mapd(:,1:c))', '--k'), title 'mean(|PD|)'
        drawnow
      end
      
      % normally converged?
      if mean(mapd(:,c)) < mean_mapd_threshold
          break
      end
      
      % relaxed convergence criterion reached
      if max(mapd(:,c-1)-mapd(:,c)) < max_mapd_diff_threshold && c > min_nsamples
          break
      end
      
      % max iters reached
      if c > max_iters
        warning('Maximum number of iterations exceeded.');
        break
      end
      
    end
    
  end
  fprintf('\n')
    
end

% Sample from the conditional distribution p(x_masked|x_fixed,z)
% The variable names W, Z, x, and z are used as described in the paper (and suppl. material)
function [x, x_mu] = sample_x_inpainting(this, z, mask, x_fixed)
  
  npixels = prod(this.imdims);
  nfilters = this.nfilters;
  nexperts = this.nexperts;
  
  N = 0;
  for i = 1:nfilters
    expert_precision = this.experts{min(i,nexperts)}.precision;
    z{i} = expert_precision * z{i}(:);
    N = N + length(z{i});
  end
  
  ind_int = setdiff((1:npixels)' .* mask, 0);
  ind_ext = setdiff(1:npixels, ind_int);
  npixels_int = length(ind_int);
  npixels_ext = length(ind_ext);
  
  x_ext = x_fixed(ind_ext);
  
  Wt_ext = cellfun(@(F) {F(:,ind_ext)}, this.filter_matrices(1:nfilters));
  Wt_int = cellfun(@(F) {F(:,ind_int)}, this.filter_matrices(1:nfilters));
  C = vertcat(Wt_int{:})' * (spdiags(vertcat(z{:}), 0, N, N) * vertcat(Wt_ext{:}));
  
  Wt = {Wt_int{:}, speye(npixels_int)};
  z = {z{:}, this.epsilon(ones(npixels_int, 1))};
  N = N + npixels_int;
  
  Wt = vertcat(Wt{:});
  Z = spdiags(vertcat(z{:}), 0, N, N);
  
  r = randn(N, 1);
  W_sqrtZ_r = Wt' * sqrt(Z) * r;
  W_Z_Wt = Wt' * Z * Wt;
  
  solve_sle = pml.numerical.sle_spd_solver(W_Z_Wt);
  x_int_mu = - solve_sle(C * x_ext);
  x_int = x_int_mu + solve_sle(W_sqrtZ_r);
  
  x = x_fixed; x_mu = x_fixed;
  x(ind_int) = x_int;
  x_mu(ind_int) = x_int_mu;
  
end